#pragma once
#ifndef __ia32cache_h__
#define __ia32cache_h__

#include "ia32size.h"

class ia32cache
{
public:

	enum type_
	{
		type_reserved,
		type_unified,
		type_instruction,
		type_trace,
		type_data,
		type_invalid
	};

	enum _
	{
		level_TLB = -1,
		associativity_Full = -1,
		block_AnySize = 0
	};

	const byte descriptor;
	const type_ type;

	const int level;

	const ia32size capacity;
	const ia32size block;

	const int associativity;

	ia32cache (byte descriptor, type_ type, int level, ia32size capacity, ia32size block, int associativity) :
		descriptor(descriptor),
		type(type),
		level(level),
		capacity(capacity),
		block(block),
		associativity(associativity)
	{
	}

	operator const string () const
	{
		if (descriptor == 0x00)
			return string("0x00: reserved");
		else
		{
			char b[256];

			if (level != level_TLB)
				sprintf(b, "0x%02x: L%d %11s$, Size(%6s), Block(%6s), Associativity(%5s)", 
					descriptor, 
					level, 
					type_text(), 
					((string)capacity).c_str(),
					((string)block).c_str(),
					associativity_text().c_str());
			else
				sprintf(b, "0x%02x: TLB %11s, Entries(%3I64d), PageSize(%s), Associativity(%5s)",
					descriptor,
					type_text(),
					(uint64)capacity,
					(uint64)block == 0 ? "*" : ((string)block).c_str(),
					associativity_text().c_str());

			return string(b);
		}
	}

protected:

	const char * type_text () const
	{
		static const char *text[] = 
		{
			"reserved",
			"unified",
			"instruction",
			"trace",
			"data"
		};

		if (type < type_invalid)
			return text[type];
		else
			return text[type_reserved];
	}

	const string associativity_text () const 
	{
		if (associativity == associativity_Full)
			return string("Full");
		else
		{
			char b[16];

			sprintf(b, "%d-way", associativity);

			return string(b);
		}
	}
};

extern const ia32cache ia32caches[];

const ia32cache &_ia32cache (byte descriptor);

#endif