#pragma once
#ifndef __ia32ring0_h__
#define __ia32ring0_h__

#include "ia32detect.h"
#include "ia32driver.h"
#include "ia32error.h"

class ia32ring0
{
	HANDLE h;

public:

	ia32ring0 ()
	{
		ia32detect ia32;

		if (ia32.version.Family < 5 || !ia32.feature.MSR)
			throw ia32error(ia32error::err_ring0_cpu);

		h = CreateFile("\\\\.\\ia32cpu", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

		if (h == INVALID_HANDLE_VALUE)
			throw ia32error(ia32error::err_ring0_create);
	}

	uint64 rdmsr (uint32 i) const 
	{
		uint64 d;

		DWORD l;

		if (DeviceIoControl(h, IOCTL_IA32CPU_READ_MSR, &i, sizeof(i), &d, sizeof(d), &l, NULL))
			if (l != sizeof(d))
				throw ia32error(ia32error::err_ring0_size);
			else
				return d;
		else
			throw ia32error(ia32error::err_ring0_ioctl);
	};

	void wrmsr (uint32 i, uint64 d) const
	{
#		pragma pack(push, 1)

		struct
		{
			uint32	i;
			uint64	d;
		}	b = {i, d};

#		pragma pack(pop)

		DWORD l;

		if (!DeviceIoControl(h, IOCTL_IA32CPU_WRITE_MSR, &b, sizeof(b), NULL, 0, &l, NULL))
			throw ia32error(ia32error::err_ring0_ioctl);
	}

	~ia32ring0 ()
	{
		if (!CloseHandle(h))
			throw ia32error(ia32error::err_ring0_close);
	}
};

#endif