#pragma once

#ifndef __p6counter_h__
#define __p6counter_h__

#include "ia32counter.h"
#include "ia32detect.h"

class p6counter: public ia32counter
{
public:

	enum event_
	{
		// Data Cache Unit (DCU)
		DCU_MEMORY_REFERENCE	= 0x43, // DATA_MEM_REFS
		DCU_LINES_IN			= 0x45,
		DCU_M_LINES_IN			= 0x46,
		DCU_M_LINES_OUT			= 0x47,
		DCU_MISS_OUTSTANDING	= 0x48,

		// Instruction Fetch Unit (IFU)
		IFU_IFETCH				= 0x80,
		IFU_IFETCH_MISS			= 0x81,
		IFU_TLB_MISS			= 0x85, // ITLB_MISS
		IFU_MEMORY_STALL		= 0x86,
		IFU_ILD_STALL			= 0x87, // ILD_STALL

		// L2 Cache
		L2_IFETCH				= 0x28,
		L2_LOADS				= 0x29, // L2_LD
		L2_STORES				= 0x2A, // L2_ST
		L2_LINES_IN				= 0x24,
		L2_LINES_OUT			= 0x26,
		L2_M_LINES_IN			= 0x25,
		L2_M_LINES_OUT			= 0x27,
		L2_REQUEST				= 0x2E, // L2_RQSTS
		L2_ADDRESS_STROBE		= 0x21, // L2_ADS
		L2_DATA_BUS_BUSY		= 0x22, // L2_DBUS_BUSY
		L2_DATA_BUS_BUSY_READ	= 0x23, // L2_DBUS_BUSY_RD

		// External Bus Logic (EBL)
		EBL_DATA_READY			= 0x62, // BUS_DRDY_CLOCKS
		EBL_LOCK				= 0x63, // BUS_LOCK_CLOCKS
		EBL_REQ_OUTSTANDING		= 0x60, // BUS_REQ_OUTSTANDING
		EBL_TRANS_BURST_READ	= 0x65, // BUS_TRAN_BRD
		EBL_TRANS_READ_OWNER	= 0x66, // BUS_TRAN_RFO
		EBL_TRANS_WRITEBACK		= 0x67, // BUS_TRANS_WB
		EBL_TRANS_IFETCH		= 0x68, // BUS_TRAN_IFETCH
		EBL_TRANS_INVALIDATE	= 0x69, // BUS_TRAN_INVAL
		EBL_TRANS_PARTIAL_WRITE	= 0x6A, // BUS_TRAN_PWR
		EBL_TRANS_PARTIAL		= 0x6B, // BUS_TRANS_P
		EBL_TRANS_IO			= 0x6C, // BUS_TRANS_IO
		EBL_TRANS_DEFERRED		= 0x6D, // BUS_TRAN_DEF
		EBL_TRANS_BURST			= 0x6E, // BUS_TRAN_BURST
		EBL_TRANS_ANY			= 0x70, // BUS_TRAN_ANY
		EBL_TRANS_MEMORY		= 0x6F, // BUS_TRAN_MEM
		EBL_DATA_RECEIVE		= 0x64, // BUS_DATA_RCV
		EBL_DRIVE_BNR			= 0x61, // BUS_BNR_DRV
		EBL_DRIVE_HIT			= 0x7A, // BUS_HIT_DRV
		EBL_DRIVE_HITM			= 0x7B, // BUS_HITM_DRV
		EBL_SNOOP_STALL			= 0x7E, // BUS_SNOOP_STALL

		// Floating-Point Unit (FPU)
		FPU_FLOPS_RETIRED		= 0xC1, // FLOPS,			Counter 0 only
		FPU_FLOPS_EXECUTED		= 0x10, // FP_COMP_OPS_EXE,	Counter 0 only
		FPU_ASSIST				= 0x11, // FP_ASSIST,		Counter 1 only
		FPU_MUL					= 0x12, // MUL,				Counter 1 only
		FPU_DIV					= 0x13, // DIV,				Counter 1 only
		FPU_DIV_BUSY			= 0x14, // CYCLES_DIV_BUSY,	Counter 0 only

		// Memory Ordering (MO)
		MO_LOAD_BLOCKED			= 0x03, // LD_BLOCKS
		MO_STORE_BUFFER_DRAIN	= 0x04, // SB_DRAINS
		MO_MISALLIGNMENT		= 0x05, // MISALIGN_MEM_REF
		SSE_PREFETCH_DISPATCHED = 0x07, // EMON_KNI_PREF_DISPATCHED
		SSE_PREFETCH_MISS		= 0x4B, // EMON_KNI_PREF_MISS

		// Instruction Decoding and Retirement (IDR)
		IDR_INSTRUCTION_RETIRED = 0xC0, // INST_RETIRED
		IDR_UOP_RETIRED			= 0xC2, // UOPS_RETIRED
		IDR_INSTRUCTION_DECODED	= 0xD0, // INST_DECODED
		SSE_INSTRUCTION_RETIRED	= 0xD8, // EMON_KNI_INST_RETIRED
		SSE_COMPUTATION_RETIRED	= 0xD9, // EMON_KNI_COMP_INST_RET

		// Interrupts (INT)
		INT_HW_RECEIVED			= 0xC8, // HW_INT_RX
		INT_MASKED				= 0xC6, // CYCLES_INT_MASKED
		INT_PENDING_AND_MASKED	= 0xC7, // CYCLES_INT_PENDING_AND_MASKED

		// Branches (BR)
		BR_INSTRUCTION_RETIRED	= 0xC4, // BR_INST_RETIRED
		BR_MISSPREDICT_RETIRED	= 0xC5, // BR_MISS_PRED_RETIRED
		BR_TAKEN_RETIRED		= 0xC6,
		BR_MISSPREDICT_TAKEN_RETIRED = 0xC7, // BR_MISS_PRED_TAKEN_RET
		BR_INSTRUCTION_DECODED	= 0xE0, // BR_INST_DECODED
		BR_BTB_MISS				= 0xE2, // BTB_MISSES
		BR_BOGUS				= 0xE4,
		BR_BACLEAR				= 0xE6, // BARCLEARS

		// Stalls (STALL)
		STALL_RESOURCE			= 0xA2, // RESOURCE_STALLS
		STALL_PARTIAL			= 0xD2, // PARTIAL_RAT_STALLS

		// Multimedia Extensions (MMX)
		MMX_INSTRUCTION_EXECUTE	= 0xB0, // MMX_INSTR_EXEC
		MMX_SATURATING_EXECUTE	= 0xB1, // MMX_SAT_INSTR_EXEC
		MMX_UOP_EXECUTE			= 0xB2, // MMX_UPOS_EXEC
		MMX_TYPE_EXECUTE		= 0xB3, // MMX_INSTR_TYPE_EXEC
		MMX_FPU_TRANSITION		= 0xCC, // FP_MMX_TRANS
		MMX_ASSIST				= 0xCD,
		MMX_INSTRUCTION_RETIRED	= 0xCE, // MMX_INSTR_RET

		// Segment Register Renaming (SRR)
		SRR_STALL				= 0xD4, // SEG_RENAME_STALLS
		SRR_COUNT				= 0xD5, // SEG_REG_RENAME
		SRR_COUNT_RETIRED		= 0xD6, // RET_SEG_RENAMES

		SEGMENT_REGISTER_LOADS	= 0x06, // SEGMENT_REG_LOADS
		CPU_CLOCKS_UNHALTED		= 0x79  // CPU_CLK_UNHALTED
	};

	enum mask_
	{
		NONE					= 0x0,

		L2_M					= 0x8,
		L2_E					= 0x4,
		L2_S					= 0x2,
		L2_I					= 0x1,
		L2_MESI					= 0xF,

		EBL_SELF				= 0x00,
		EBL_ANY					= 0x20,

		SSE_PREFETCH_NTA		= 0x00,
		SSE_PREFETCH_T1			= 0x01,
		SSE_PREFETCH_T2			= 0x02,
		SSE_WEAKLY_ORDERED_STORES = 0x03,

		SSE_PACKED_AND_SCALAR	= 0x00,
		SSE_SCALAR				= 0x01,

		MMX_PACKED_MULTIPLY		= 0x01,
		MMX_PACKED_SHIFT		= 0x02,
		MMX_PACK				= 0x04,
		MMX_UNPACK				= 0x08,
		MMX_PACKED_LOGICAL		= 0x10,
		MMX_PACKED_ARITHMETIC	= 0x20,
		MMX_ANY					= 0x3F,

		MMX_TO_FPU				= 0x0,
		MMX_FROM_FPU			= 0x1,

		SRR_ES					= 0x1,
		SRR_DS					= 0x2,
		SRR_FS					= 0x4,
		SRR_GS					= 0x8,
		SRR_ANY					= 0xF
	};

#	pragma pack(push, 1)

	struct
	{
		bit	event		: 8;
		bit	mask		: 8;
		bit	ring123		: 1;
		bit	ring0		: 1;
		bit	edge		: 1;
		bit	pin			: 1;
		bit	int_		: 1;
		bit	reserved	: 1;
		bit	enable		: 1;
		bit	invert		: 1;
		bit	count		: 8;
	} config;

#	pragma pack(pop)

	ia32ring0 r0;

	p6counter (event_ event, mask_ mask = NONE, byte count = 0, bool invert = false) : 
		ia32counter(2)
	{
		config.event = event;
		config.mask = mask;
		config.count = count;

		config.invert = invert;

		config.ring0 = 0;
		config.ring123 = 1;

		config.edge = 0;
		config.pin = 0;
		config.int_ = 0;

		config.reserved = 0;

		config.enable = 1;

		ia32detect ia32;

		if (ia32.version.Family != 6)
			throw ia32error(ia32error::err_counter_family);

		if (!ia32.feature.MMX && (
			event == MMX_INSTRUCTION_EXECUTE ||
			event == MMX_SATURATING_EXECUTE ||
			event == MMX_UOP_EXECUTE ||
			event == MMX_TYPE_EXECUTE ||
			event == MMX_FPU_TRANSITION ||
			event == MMX_ASSIST ||
			event == MMX_INSTRUCTION_RETIRED))
			throw ia32error(ia32error::err_counter_MMX);

		if (!ia32.feature.SSE && (
			event == SSE_INSTRUCTION_RETIRED ||
			event == SSE_COMPUTATION_RETIRED ||
			event == SSE_PREFETCH_DISPATCHED || 
			event == SSE_PREFETCH_MISS))
			throw ia32error(ia32error::err_counter_SSE);

		if (index != 0 && (
			event == FPU_FLOPS_RETIRED ||
			event == FPU_FLOPS_EXECUTED ||
			event == FPU_DIV_BUSY) ||
			index != 1 && (
			event == FPU_ASSIST ||
			event == FPU_MUL ||
			event == FPU_DIV))
			throw ia32error(ia32error::err_counter_counter);

		r0.wrmsr(0x186 + index, 0);
		r0.wrmsr(0xC1 + index, 0);
		r0.wrmsr(0x186 + index, *(uint32 *)&config);
	}

	operator const uint64 () const
	{
		return r0.rdmsr(0xC1 + index) & 0xFFFFFFFFFF;

		/*
		uint32 i = index;
		uint64 d;

		__asm
		{
			mov	eax, i;
			rdpmc;
			mov	dword ptr d + 0, eax;
			mov dword ptr d + 4, edx;
		}

		return d;
		*/
	}
};

#endif